/*
 * player_client_example.cc
 *
 * example Player/Stage client program
 * 
 * Subcribes to robot position and laser data,
 * asks user for x,y location to drive robot to.
 * Drives directly to new location without any
 * obstacle avoidance.
 * 
 */

#include <libplayerc++/playerc++.h>
#include <iostream>

#include "args.h"

int main(int argc, char **argv)
{
	// Interpret command line args into gHostname, gPort, other global options
	parse_args(argc,argv);

	// we throw exceptions on creation if we fail
	try
	{
		// Use player namespace to get easy access to player functions
		using namespace PlayerCc;
		// Use standard namespace to get easy access to cout, etc
		using namespace std;

		// Connect to player server on the robot
		PlayerClient robot(gHostname, gPort);
		// Subsribe to position proxy on robot server
		Position2dProxy pp(&robot, gIndex);
		// Subscribe to laser proxy on robot server
		LaserProxy lp(&robot, gIndex);

		std::cout << robot << std::endl;
		
		// Set player server message behavior
		robot.SetReplaceRule(true,-1,-1,-1);
		robot.SetDataMode(gDataMode);
		
		robot.Read();
		int count = lp.GetCount();
		// wait for real data
		while( count == 0 )
		{
			robot.Read();
			count = lp.GetCount();
		}
		double angRes = lp.GetScanRes();
		
		pp.SetMotorEnable (true);
		pp.ResetOdometry();
		
		double xPos = pp.GetXPos();
		double yPos = pp.GetYPos();
		double theta = pp.GetYaw();
		
		double robotRadius = 0.25;

		double newSpeed = 0;
		double newTurnRate = 0;
		
		// Speed limits
		double maxSpeed = 0.3; // in m/s
		double maxTurnRate = 0.5; // in rad/s
		
		// goal position
		double xGoal = xPos;
		double yGoal = yPos;
		double closeEnough = 0.05;

		// go into read-think-act loop
		for(;;)
		{
			// ------- Read ------------
			// this blocks until new data comes; 10Hz by default
			robot.Read();

			// number of sensor measurements from laser
			count = lp.GetCount();
			
			// resolution of laser
			angRes = lp.GetScanRes();
			
			// position of robot
			xPos = pp.GetXPos();
			yPos = pp.GetYPos();
			theta = normalize(pp.GetYaw());

			// -------- Think ----------
			
			// Distance to goal
			double distToGoal = sqrt(pow(xGoal-xPos,2) + pow(yGoal-yPos,2));
			
			if( distToGoal < closeEnough )
			{
				pp.SetSpeed( 0.0, 0.0 );
				cout<< "Input new goal position" << endl;
				cout<< "Cur X = " << xPos << ", new? ";
				cin>> xGoal;
				cout<< "Cur Y = " << yPos << ", new? ";
				cin>> yGoal;
				continue;
			}
			
			// Angle towards the goal in robot's frame
			double angleToGoal = normalize(atan2(yGoal-yPos,xGoal-xPos) - theta);
			
			cout<< "Goal is " << distToGoal << "m away at angle of ";
			cout << rtod(angleToGoal) << "deg" << endl;
			
			double minObsDist = lp.GetMaxRange();
			
			// loop through laser distances
			for( int j = 0; j < count; j++ )
			{
				// Angle in robot's frame for the jth measurement
				double angle = angRes*(j - count/2);
				// Distance to closest obstacle at this angle
				double distance = lp[j];
				
				if( distance < minObsDist )
				{
					minObsDist = distance;
				}
				
			}
			
			// In robot's frame
			double desiredHeading = angleToGoal;
			
			// calculate speed and turn rate
			newSpeed = maxSpeed*limit(2*(distToGoal),0.0,1.0);
			newSpeed *= limit(minObsDist,0.0,1.0);
			newSpeed *= limit(1.1 - abs(desiredHeading),0.0,1.0);
			newTurnRate = desiredHeading;
			
			// Limit speeds based on defined maximums
			newSpeed = limit( newSpeed, -maxSpeed/2, maxSpeed );
			newTurnRate = limit(newTurnRate,-maxTurnRate,maxTurnRate);
	
			cout << "speed: " << newSpeed << " m/s "
				  << "turn: " << newTurnRate << " rad/s"
				  << endl << endl;
				  
			// ---------- Act -------------
			// write commands to robot
			pp.SetSpeed(newSpeed, newTurnRate);
    	}
	}
	catch (PlayerCc::PlayerError e)
	{
		std::cerr << e << std::endl;
		return -1;
	}
}
